/*!
  @file rx_sph_cpu.cpp
	
  @brief SPH@(GPU)̎
 
  @author Makoto Fujisawa
  @date   2008-10,2011-06
*/
// FILE --rx_sph_cpu.cpp--

#pragma comment(lib, "gsl.lib")
#pragma comment(lib, "gslcblas.lib")

//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include "rx_sph.h"

#include "rx_cu_funcs.cuh"
#include "rx_cu_common.cuh"

#include "rx_pcube.h"

#include <gsl/gsl_linalg.h>


//-----------------------------------------------------------------------------
// nbVlɂ\[gp̍\
//-----------------------------------------------------------------------------
struct rxHashSort
{
	uint hash;
	uint value;
};

/*!
 * nbVl̔r֐
 * @param[in] left,right rl
 * @return left < right
 */
inline bool LessHash(const rxHashSort &left, const rxHashSort &right)
{
	return left.hash < right.hash;
}



//-----------------------------------------------------------------------------
// rxSPHNX̎
//-----------------------------------------------------------------------------
/*!
 * RXgN^
 * @param[in] use_opengl VBOgptO
 */
rxSPH::rxSPH(bool use_opengl) : 
	rxParticleSystemBase(use_opengl), 
	m_hNrm(0), 
	m_hFrc(0),
	m_hDens(0), 
	m_hPres(0), 
	m_hUpPos(0), 
	m_hPosW(0), 
	m_hEigen(0), 
	m_hRMatrix(0), 
	m_hG(0), 
	m_hVrts(0), 
	m_hTris(0), 
	m_pBoundary(0)
{
	m_v3Gravity = Vec3(0.0, -9.82, 0.0);

	m_fViscosity = 6.0f;
	m_fGasStiffness = 3.0f;
	m_fBuoyancy = 0.0f;

	m_fDamping = 0.0;
	m_fRestitution = 0.0;

	m_bUseVorticity = false;
	m_bGridVelocity = false;
	m_bCalNormal = false;
	m_bCalAnisotropic = false;

	m_hCellData.uNumCells = 0;
	m_hCellData.hSortedIndex = 0;
	m_hCellData.hGridParticleHash = 0;
	m_hCellData.hCellStart = 0;
	m_hCellData.hCellEnd = 0;

	m_hCellData.hSortedPolyIdx = 0;
	m_hCellData.hGridPolyHash = 0;
	m_hCellData.hPolyCellStart = 0;
	m_hCellData.hPolyCellEnd = 0;
	m_hCellData.uNumPolyHash = 0;

	m_uNumParticles = 0;
	m_iNumTris = 0;

	m_iColorType = RX_RAMP;
}

/*!
 * fXgN^
 */
rxSPH::~rxSPH()
{
	Finalize();
}


/*!
 * V~[V̏
 * @param[in] max_particles őp[eBN
 * @param[in] boundary_ext Ȇ傫(eӂ̒1/2)
 * @param[in] dens x
 * @param[in] mass p[eBN̎
 * @param[in] kernel_particle Lahȓ̃p[eBN
 */
void rxSPH::Initialize(const rxSPHEnviroment &env)
{
	// MARK:Initialize
	RXCOUT << "[rxSPH::Initialize]" << endl;

	m_fInitDens        = env.dens;
	m_fMass            = env.mass;
	m_iKernelParticles = env.kernel_particles;

	RXREAL volume = env.max_particles*m_fMass/m_fInitDens;

	m_fEffectiveRadius = pow(((3.0*m_iKernelParticles*volume)/(4.0*env.max_particles*RX_PI)), 1.0/3.0);
	m_fParticleRadius = 0.5f*m_fEffectiveRadius;

	RXCOUT << "particle : " << endl;
	RXCOUT << " n_max = " << env.max_particles << endl;
	RXCOUT << " h = " << m_fEffectiveRadius << endl;
	RXCOUT << " r = " << m_fParticleRadius << endl;
	RXCOUT << " dens = " << m_fInitDens << endl;
	RXCOUT << " mass = " << m_fMass << endl;
	RXCOUT << " kernel_particles = " << m_iKernelParticles << endl;
	RXCOUT << " volume = " << volume << endl;

	RXREAL h = m_fEffectiveRadius;
	RXREAL r = m_fParticleRadius;

	//
	// Eݒ
	//
	m_pBoundary = new rxSolidBox(env.boundary_cen-env.boundary_ext, env.boundary_cen+env.boundary_ext, -1);
	//m_pBoundary = new rxSolidSphere(Vec3(0.0, 0.1, 0.0), 0.25, -1);
	
	// V~[V̑傫
	m_v3EnvMin = m_pBoundary->GetMin();
	m_v3EnvMax = m_pBoundary->GetMax();
	RXCOUT << "simlation range : " << m_v3EnvMin << " - " << m_v3EnvMax << endl;

	Vec3 world_size = m_v3EnvMax-m_v3EnvMin;
	Vec3 world_origin = m_v3EnvMin;
	
	double expansion = 0.01;
	world_origin -= 0.5*expansion*world_size;
	world_size *= (1.0+expansion); // V~[VŜ𕢂悤ɐݒ

	m_v3EnvMin = world_origin;
	m_v3EnvMax = world_origin+world_size;

	// Zݒ
	setupCells(m_v3EnvMin, m_v3EnvMax, h);
	
	RXCOUT << "grid for nn search : " << endl;
	RXCOUT << "  size   : " << m_iGridSize[0] << "x" << m_iGridSize[1] << "x" << m_iGridSize[2] << endl;
	RXCOUT << "  num    : " << m_hCellData.uNumCells << endl;
	RXCOUT << "  origin : " << m_v3EnvMin << endl;
	RXCOUT << "  width  : " << m_fCellWidth[0] << "x" << m_fCellWidth[1] << "x" << m_fCellWidth[2] << endl;

	m_uNumArdGrid = (int)(m_fEffectiveRadius/m_fCellWidth[0])+1;
	RXCOUT << "  numArdGrid : " << m_uNumArdGrid << endl;

	// J[l֐̒萔
	m_fWpoly6	=  315.0/(64.0*RX_PI*pow(h, (RXREAL)9.0));
	m_fGWpoly6	= -945.0/(32.0*RX_PI*pow(h, (RXREAL)9.0));
	m_fLWpoly6	= -945.0/(32.0*RX_PI*pow(h, (RXREAL)9.0));

	m_fWspiky	=  15.0/(RX_PI*pow(h, (RXREAL)6.0));
	m_fGWspiky	= -45.0/(RX_PI*pow(h, (RXREAL)6.0));
	m_fLWspiky	= -90.0/(RX_PI*pow(h, (RXREAL)6.0));

	m_fWvisc	= 15.0/(2.0*RX_PI*pow(h, (RXREAL)3.0));
	m_fGWvisc	= 15.0/(2.0*RX_PI*pow(h, (RXREAL)3.0));
	m_fLWvisc	= 45.0/(RX_PI*pow(h, (RXREAL)6.0));

	m_uNumParticles = 0;

	Allocate(env.max_particles);
}

/*!
 * ̊m
 *  - őp[eBNŊm
 * @param[in] max_particles őp[eBN
 */
void rxSPH::Allocate(int max_particles)
{
	// MARK:Allocate
	assert(!m_bInitialized);

	//m_uNumParticles = max_particles;
	m_uMaxParticles = max_particles;
	
	unsigned int size  = m_uMaxParticles*DIM;
	unsigned int size1 = m_uMaxParticles;
	unsigned int mem_size  = sizeof(RXREAL)*size;
	unsigned int mem_size1 = sizeof(RXREAL)*size1;

	//
	// m
	//
	m_hPos = new RXREAL[size];
	m_hVel = new RXREAL[size];
	m_hNrm = new RXREAL[size];
	m_hFrc = new RXREAL[size];
	memset(m_hPos, 0, mem_size);
	memset(m_hVel, 0, mem_size);
	memset(m_hNrm, 0, mem_size);
	memset(m_hFrc, 0, mem_size);

	m_hDens = new RXREAL[size1];
	m_hPres = new RXREAL[size1];
	memset(m_hDens, 0, mem_size1);
	memset(m_hPres, 0, mem_size1);

	// Anisotropic kernel
	m_hUpPos = new RXREAL[size];
	m_hPosW = new RXREAL[size];
	m_hEigen = new RXREAL[m_uMaxParticles*3];
	m_hRMatrix = new RXREAL[m_uMaxParticles*9];
	m_hG = new RXREAL[m_uMaxParticles*9];

	m_vNeighs.resize(m_uMaxParticles);

	if(m_bUseOpenGL){
		m_posVBO = createVBO(mem_size);	
		m_colorVBO = createVBO(m_uMaxParticles*4*sizeof(RXREAL));

		SetColorVBO(RX_RAMP);
	}

	// CPUvZpZ
	m_hCellData.hSortedIndex = new uint[m_uMaxParticles];
	m_hCellData.hGridParticleHash = new uint[m_uMaxParticles];
	m_hCellData.hCellStart = new uint[m_hCellData.uNumCells];
	m_hCellData.hCellEnd = new uint[m_hCellData.uNumCells];

	m_hCellData.hPolyCellStart = new uint[m_hCellData.uNumCells];
	m_hCellData.hPolyCellEnd = new uint[m_hCellData.uNumCells];

	m_bInitialized = true;
}

/*!
 * mۂ̉
 */
void rxSPH::Finalize(void)
{
	assert(m_bInitialized);

	// 
	if(m_hPos) delete [] m_hPos;
	if(m_hVel) delete [] m_hVel;
	if(m_hNrm) delete [] m_hNrm;
	if(m_hFrc) delete [] m_hFrc;

	if(m_hDens) delete [] m_hDens;
	if(m_hPres) delete [] m_hPres;

	// Anisotoropic kernel
	if(m_hUpPos) delete [] m_hUpPos;
	if(m_hPosW) delete [] m_hPosW;
	if(m_hEigen) delete [] m_hEigen;
	if(m_hRMatrix) delete [] m_hRMatrix;
	if(m_hG) delete [] m_hG;

	m_vNeighs.clear();

	if(m_bUseOpenGL){
		glDeleteBuffers(1, (const GLuint*)&m_posVBO);
		glDeleteBuffers(1, (const GLuint*)&m_colorVBO);
	}

	if(m_hVrts) delete [] m_hVrts;
	if(m_hTris) delete [] m_hTris;

	if(m_hCellData.hSortedIndex) delete [] m_hCellData.hSortedIndex;
	if(m_hCellData.hGridParticleHash) delete [] m_hCellData.hGridParticleHash;
	if(m_hCellData.hCellStart) delete [] m_hCellData.hCellStart;
	if(m_hCellData.hCellEnd) delete [] m_hCellData.hCellEnd;

	if(m_hCellData.hSortedPolyIdx) delete [] m_hCellData.hSortedPolyIdx;
	if(m_hCellData.hGridPolyHash) delete [] m_hCellData.hGridPolyHash;
	if(m_hCellData.hPolyCellStart) delete [] m_hCellData.hPolyCellStart;
	if(m_hCellData.hPolyCellEnd) delete [] m_hCellData.hPolyCellEnd;

	if(m_pBoundary) delete m_pBoundary;

	int num_solid = (int)m_vSolids.size();
	for(int i = 0; i < num_solid; ++i){
		delete m_vSolids[i];
	}
	m_vSolids.clear();	

	m_uNumParticles = 0;
	m_uMaxParticles = 0;
}


/*!
 * SPH1Xebvi߂
 * @param[in] dt ԃXebv
 * @retval ture  vZ
 * @retval false őXebv𒴂Ă܂
 */
bool rxSPH::Update(RXREAL dt)
{
	// MARK:Update : ԍXV
	assert(m_bInitialized);

	Vec3 v_half, acc_f;

	static bool init = false;

	for(uint j = 0; j < m_solverIterations; ++j){
		// Zɗqo^
		setObjectToCell(m_hPos);

		// ߖTqT
		searchNeighbors(m_hPos, m_fEffectiveRadius);

		RXTIMER("cell construction");

		// xvZ
		calDensity();

		// ͍,ŠvZ
		calForce();

		RXTIMER("force calculation");

		// ʒuEx̍XV
		integrate(m_hPos, m_hVel, m_hFrc, m_hPos, m_hVel, dt);

		RXTIMER("update position");
	}


	SetArrayVBO(RX_POSITION, m_hPos, 0, m_uNumParticles);

	SetColorVBO(m_iColorType);

	RXTIMER("color(vbo)");

	return true;
}

/*!
 * p[eBNf[^̎擾
 * @return p[eBNf[^̃foCX|C^
 */
RXREAL* rxSPH::GetParticle(void)
{
	return m_hPos;
}

/*!
 * p[eBNf[^̎擾
 * @return p[eBNf[^̃foCX|C^
 */
RXREAL* rxSPH::GetParticleDevice(void)
{
	if(!m_uNumParticles) return 0;

	RXREAL *dPos = 0;
	CuAllocateArray((void**)&dPos, m_uNumParticles*4*sizeof(RXREAL));

	CuCopyArrayToDevice(dPos, m_hPos, 0, m_uNumParticles*4*sizeof(RXREAL));

	return dPos;
}


inline float RxPythag(const float a, const float b)
{
	float absa = abs(a), absb = abs(b);
	return (absa > absb ? absa*(float)sqrt((double)(1.0+(absb/absa)*(absb/absa))) :
		   (absb == 0.0 ? 0.0 : absb*(float)sqrt((double)(1.0+(absa/absb)*(absa/absb)))));
}

void RxSVDecomp3(float w[3], float u[9], float v[9], float eps)
{
	bool flag;
	int i, its, j, jj, k, l, nm;
	float anorm, c, f, g, h, s, scale, x, y, z;
	float rv1[3];
	g = scale = anorm = 0.0;
	for(i = 0; i < 3; ++i){
		l = i+2;
		rv1[i] = scale*g;
		g = s = scale = 0.0;
		for(k = i; k < 3; ++k) scale += abs(u[k*3+i]);
		if(scale != 0.0){
			for(k = i; k < 3; ++k){
				u[k*3+i] /= scale;
				s += u[k*3+i]*u[k*3+i];
			}
			f = u[i*3+i];
			g = -RX_SIGN2(sqrt(s), f);
			h = f*g-s;
			u[i*3+i] = f-g;
			for(j = l-1; j < 3; ++j){
				for(s = 0.0, k = i; k < 3; ++k) s += u[k*3+i]*u[k*3+j];
				f = s/h;
				for(k = i; k < 3; ++k) u[k*3+j] += f*u[k*3+i];
			}
			for(k = i; k < 3; ++k) u[k*3+i] *= scale;
		}

		w[i] = scale*g;
		g = s = scale = 0.0;
		if(i+1 <= 3 && i+1 != 3){
			for(k = l-1; k < 3; ++k) scale += abs(u[i*3+k]);
			if(scale != 0.0){
				for(k = l-1; k < 3; ++k){
					u[i*3+k] /= scale;
					s += u[i*3+k]*u[i*3+k];
				}
				f = u[i*3+l-1];
				g = -RX_SIGN2(sqrt(s), f);
				h = f*g-s;
				u[i*3+l-1] = f-g;
				for(k = l-1; k < 3; ++k) rv1[k] = u[i*3+k]/h;
				for(j = l-1; j < 3; ++j){
					for(s = 0.0,k = l-1; k < 3; ++k) s += u[j*3+k]*u[i*3+k];
					for(k = l-1; k < 3; ++k) u[j*3+k] += s*rv1[k];
				}
				for(k = l-1; k < 3; ++k) u[i*3+k] *= scale;
			}
		}
		anorm = RX_MAX(anorm, (abs(w[i])+abs(rv1[i])));
	}
	for(i = 2; i >= 0; --i){
		if(i < 2){
			if(g != 0.0){
				for(j = l; j < 3; ++j){
					v[j*3+i] = (u[i*3+j]/u[i*3+l])/g;
				}
				for(j = l; j < 3; ++j){
					for(s = 0.0, k = l; k < 3; ++k) s += u[i*3+k]*v[k*3+j];
					for(k = l; k < 3; ++k) v[k*3+j] += s*v[k*3+i];
				}
			}
			for(j = l; j < 3; ++j) v[i*3+j] = v[j*3+i] = 0.0;
		}
		v[i*3+i] = 1.0;
		g = rv1[i];
		l = i;
	}
	for(i = 2; i >= 0; --i){
		l = i+1;
		g = w[i];
		for(j = l; j < 3; ++j) u[i*3+j] = 0.0;
		if(g != 0.0){
			g = 1.0/g;
			for(j = l; j < 3; ++j){
				for(s = 0.0, k = l; k < 3; ++k) s += u[k*3+i]*u[k*3+j];
				f = (s/u[i*3+i])*g;
				for(k = i; k < 3; ++k) u[k*3+j] += f*u[k*3+i];
			}
			for(j = i; j < 3; ++j) u[j*3+i] *= g;
		}
		else{
			for(j = i; j < 3; ++j) u[j*3+i] = 0.0;
		}
		++u[i*3+i];
	}
	for(k = 2; k >= 0; --k){
		for(its = 0; its < 30; ++its){
			flag = true;
			for(l = k; l >= 0; --l){
				nm = l-1;
				if(l == 0 || abs(rv1[l]) <= eps*anorm){
					flag = false;
					break;
				}
				if(abs(w[nm]) <= eps*anorm) break;
			}
			if(flag){
				c = 0.0;
				s = 1.0;
				for(i = l; i < k+1; ++i){
					f = s*rv1[i];
					rv1[i] = c*rv1[i];
					if(abs(f) <= eps*anorm) break;
					g = w[i];
					h = RxPythag(f, g);
					w[i] = h;
					h = 1.0/h;
					c = g*h;
					s = -f*h;
					for(j = 0; j < 3; ++j){
						y = u[j*3+nm];
						z = u[j*3+i];
						u[j*3+nm] = y*c+z*s;
						u[j*3+i] = z*c-y*s;
					}
				}
			}
			z = w[k];
			if(l == k){
				if(z < 0.0){
					w[k] = -z;
					for(j = 0; j < 3; ++j) v[j*3+k] = -v[j*3+k];
				}
				break;
			}
			if(its == 29){
				printf("no convergence in 30 svdcmp iterations");
				return;
			}
			x = w[l];
			nm = k-1;
			y = w[nm];
			g = rv1[nm];
			h = rv1[k];
			f = ((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y);
			g = RxPythag(f, 1.0f);
			f = ((x-z)*(x+z)+h*((y/(f+RX_SIGN2(g, f)))-h))/x;
			c = s = 1.0;
			for(j = l; j <= nm; ++j){
				i = j+1;
				g = rv1[i];
				y = w[i];
				h = s*g;
				g = c*g;
				z = RxPythag(f, h);
				rv1[j] = z;
				c = f/z;
				s = h/z;
				f = x*c+g*s;
				g = g*c-x*s;
				h = y*s;
				y *= c;
				for(jj = 0; jj < 3; ++jj){
					x = v[jj*3+j];
					z = v[jj*3+i];
					v[jj*3+j] = x*c+z*s;
					v[jj*3+i] = z*c-x*s;
				}
				z = RxPythag(f, h);
				w[j] = z;
				if(z){
					z = 1.0/z;
					c = f*z;
					s = h*z;
				}
				f = c*g+s*y;
				x = c*y-s*g;
				for(jj = 0; jj < 3; ++jj){
					y = u[jj*3+j];
					z = u[jj*3+i];
					u[jj*3+j] = y*c+z*s;
					u[jj*3+i] = z*c-y*s;
				}
			}
			rv1[l] = 0.0;
			rv1[k] = f;
			w[k] = x;
		}
	}

	// reorder
	int inc = 1;
	float sw;
	float su[3], sv[3];

	do{
		inc *= 3;
		inc++; 
	}while(inc <= 3);

	do{
		inc /= 3;
		for(i = inc; i < 3; ++i){
			sw = w[i];
			for(k = 0; k < 3; ++k) su[k] = u[k*3+i];
			for(k = 0; k < 3; ++k) sv[k] = v[k*3+i];
			j = i;
			while (w[j-inc] < sw){
				w[j] = w[j-inc];
				for(k = 0; k < 3; ++k) u[k*3+j] = u[k*3+j-inc];
				for(k = 0; k < 3; ++k) v[k*3+j] = v[k*3+j-inc];
				j -= inc;
				if (j < inc) break;
			}
			w[j] = sw;
			for(k = 0; k < 3; ++k) u[k*3+j] = su[k];
			for(k = 0; k < 3; ++k) v[k*3+j] = sv[k];

		}
	}while(inc > 1);

	for(k = 0; k < 3; ++k){
		s = 0;
		for(i = 0; i < 3; ++i) if(u[i*3+k] < 0.) s++;
		for(j = 0; j < 3; ++j) if(v[j*3+k] < 0.) s++;
		if(s > 3){
			for(i = 0; i < 3; ++i) u[i*3+k] = -u[i*3+k];
			for(j = 0; j < 3; ++j) v[j*3+k] = -v[j*3+k];
		}
	}
}

/*!
 * Anisotropic kerneľvZ
 *  - J. Yu and G. Turk, Reconstructing Surfaces of Particle-Based Fluids Using Anisotropic Kernels, SCA2010. 
 */
void rxSPH::CalAnisotropicKernel(void)
{
	// MARK:CalAnisotropicKernel
	if(m_uNumParticles == 0) return;

	// VBO烁
	RXREAL r0 = m_fInitDens;
	RXREAL h0 = m_fEffectiveRadius;
	RXREAL h = g_fRadScaling*h0;

	// 2{̒TaŋߖTT
	setObjectToCell(m_hPos);
	searchNeighbors(m_hPos, h);

	// XVʒǔvZ
	for(uint i = 0; i < m_uNumParticles; ++i){
		Vec3 pos0;
		pos0[0] = m_hPos[4*i+0];
		pos0[1] = m_hPos[4*i+1];
		pos0[2] = m_hPos[4*i+2];

		// ߖTq
		Vec3 posw(0.0);
		RXREAL sumw = 0.0f;
		for(vector<rxNeigh>::iterator itr = m_vNeighs[i].begin() ; itr != m_vNeighs[i].end(); ++itr){
			int j = itr->Idx;
			if(j < 0) continue;

			Vec3 pos1;
			pos1[0] = m_hPos[4*j+0];
			pos1[1] = m_hPos[4*j+1];
			pos1[2] = m_hPos[4*j+2];

			RXREAL r = sqrt(itr->Dist2);
			
			if(r < h){
				RXREAL q = 1-r/h;
				RXREAL wij = q*q*q;
				posw += pos1*wij;
				sumw += wij;
			}
		}

		m_hPosW[DIM*i+0] = posw[0]/sumw;
		m_hPosW[DIM*i+1] = posw[1]/sumw;
		m_hPosW[DIM*i+2] = posw[2]/sumw;
		m_hPosW[DIM*i+3] = 0.0f;

		m_hUpPos[DIM*i+0] = (1-g_fLambda)*m_hPos[DIM*i+0]+g_fLambda*m_hPosW[DIM*i+0];
		m_hUpPos[DIM*i+1] = (1-g_fLambda)*m_hPos[DIM*i+1]+g_fLambda*m_hPosW[DIM*i+1];
		m_hUpPos[DIM*i+2] = (1-g_fLambda)*m_hPos[DIM*i+2]+g_fLambda*m_hPosW[DIM*i+2];

		//m_hUpPos[DIM*i+0] = m_hPos[DIM*i+0];
		//m_hUpPos[DIM*i+1] = m_hPos[DIM*i+1];
		//m_hUpPos[DIM*i+2] = m_hPos[DIM*i+2];
	}

	setObjectToCell(m_hUpPos);
	searchNeighbors(m_hUpPos, h);

	// covariance matrix̌vZ
	for(uint i = 0; i < m_uNumParticles; ++i){
		Vec3 xi;
		xi[0] = m_hUpPos[4*i+0];
		xi[1] = m_hUpPos[4*i+1];
		xi[2] = m_hUpPos[4*i+2];

		Vec3 xiw(0.0);
		RXREAL sumw = 0.0f;
		for(vector<rxNeigh>::iterator itr = m_vNeighs[i].begin() ; itr != m_vNeighs[i].end(); ++itr){
			int j = itr->Idx;
			if(j < 0) continue;

			Vec3 xj;
			xj[0] = m_hUpPos[4*j+0];
			xj[1] = m_hUpPos[4*j+1];
			xj[2] = m_hUpPos[4*j+2];

			RXREAL r = sqrt(itr->Dist2);
			
			if(r < h){
				RXREAL q = 1-r/h;
				RXREAL wij = q*q*q;
				xiw += xj*wij;
				sumw += wij;
			}
		}

		xiw /= sumw;

		matrix3x3 c;
		for(int k = 0; k < 3; ++k){
			c.e[k].x = 0.0;
			c.e[k].y = 0.0;
			c.e[k].z = 0.0;
		}

		int n = 0;
		sumw = 0.0f;
		for(vector<rxNeigh>::iterator itr = m_vNeighs[i].begin() ; itr != m_vNeighs[i].end(); ++itr){
			int j = itr->Idx;
			if(j < 0) continue;

			Vec3 xj;
			xj[0] = m_hUpPos[4*j+0];
			xj[1] = m_hUpPos[4*j+1];
			xj[2] = m_hUpPos[4*j+2];

			RXREAL r = sqrt(itr->Dist2);

			if(r < h){
				RXREAL q = 1-r/h;
				RXREAL wij = q*q*q;

				Vec3 dxj = xj-xiw;

				for(int k = 0; k < 3; ++k){
					c.e[k].x += wij*dxj[k]*dxj[0];
					c.e[k].y += wij*dxj[k]*dxj[1];
					c.e[k].z += wij*dxj[k]*dxj[2];
				}

				sumw += wij;

				n++;
			}
		}

		for(int k = 0; k < 3; ++k){
			c.e[k].x /= sumw;
			c.e[k].y /= sumw;
			c.e[k].z /= sumw;
		}

		// ْl
		gsl_vector *S = gsl_vector_alloc(3);	// 

		gsl_matrix *A = gsl_matrix_alloc(3, 3);	// ̍sA,ϊU
		gsl_matrix *V = gsl_matrix_alloc(3, 3);	// V

		gsl_vector *work = gsl_vector_alloc(3);	// Ɨ̈

		for(int k = 0; k < 3; ++k){
			gsl_matrix_set(A, k, 0, c.e[k].x);
			gsl_matrix_set(A, k, 1, c.e[k].y);
			gsl_matrix_set(A, k, 2, c.e[k].z);
		}

		gsl_linalg_SV_decomp(A, V, S, work);

		// ŗLl
		Vec3 sigma;
		for(int j = 0; j < 3; ++j){
			sigma[j] = gsl_vector_get(S, j);
		}

		// ŗLxNg(]sR)
		rxMatrix3 R;
		for(int j = 0; j < 3; ++j){
			for(int k = 0; k < 3; ++k){
				R(j, k) = gsl_matrix_get(A, j, k);
			}
		}

		gsl_matrix_free(A);
		gsl_matrix_free(V);

		gsl_vector_free(S);
		gsl_vector_free(work);

		// fobOpɒlҔ
		for(int j = 0; j < 3; ++j){
			m_hEigen[3*i+j] = sigma[j];
		}
		for(int j = 0; j < 3; ++j){
			for(int k = 0; k < 3; ++k){
				m_hRMatrix[9*i+3*j+k] = R(j, k);
			}
		}

		
		if(n > g_iNe){
			for(int j = 1; j < 3; ++j){
				sigma[j] = RX_MAX(sigma[j], sigma[0]/g_fKr);
			}
			sigma *= g_fKs;
		}
		else{
			for(int j = 0; j < 3; ++j){
				sigma[j] = g_fKn;
			}
		}


		// J[lό`sG
		rxMatrix3 G;
		for(int j = 0; j < 3; ++j){
			for(int k = 0; k < 3; ++k){
				RXREAL x = 0;
				for(int l = 0; l < 3; ++l){
					//x += m_hRMatrix[9*i+3*j+l]*m_hRMatrix[9*i+3*k+l]/m_hEigen[3*i+l];
					x += R(j, l)*R(k, l)/sigma[l];
				}
				
				G(j, k) = x;
				//m_hG[9*i+3*j+k] = x;
			}
		}

		double max_diag = -1.0e10;
		for(int j = 0; j < 3; ++j){
			for(int k = 0; k < 3; ++k){
				if(G(j, k) > max_diag) max_diag = G(j, k);
			}
		}

		for(int j = 0; j < 3; ++j){
			for(int k = 0; k < 3; ++k){
				G(j, k) /= max_diag;
			}
		}
				
		G = G.Inverse();


		for(int j = 0; j < 3; ++j){
			for(int k = 0; k < 3; ++k){
				m_hG[9*i+3*j+k] = G(j, k);
			}
		}

	}
}

/*!
 * J[lpVBO̕ҏW
 * @param[in] type F̃x[XƂ镨l
 */
void rxSPH::SetColorVBO(int type)
{
	// MRK:SetColorVBO

	switch(type){
	case RX_DENSITY:
		SetColorVBOFromArray(m_hDens, 1, false, m_fInitDens);
		break;

	case RX_PRESSURE:
		SetColorVBOFromArray(m_hPres, 1);
		break;

	case RX_RAMP:
		if(m_bUseOpenGL){
			// J[obt@ɒlݒ
			glBindBufferARB(GL_ARRAY_BUFFER, m_colorVBO);
			RXREAL *data = (RXREAL*)glMapBufferARB(GL_ARRAY_BUFFER, GL_WRITE_ONLY);
			RXREAL *ptr = data;
			for(uint i = 0; i < m_uNumParticles; ++i){
				RXREAL t = i/(RXREAL)m_uNumParticles;
#if 0
				*ptr++ = rand()/(RXREAL)RAND_MAX;
				*ptr++ = rand()/(RXREAL)RAND_MAX;
				*ptr++ = rand()/(RXREAL)RAND_MAX;
#else
				RX_COLOR_RAMP(t, ptr);
				ptr += 3;
#endif
				*ptr++ = 1.0f;
			}
			glUnmapBufferARB(GL_ARRAY_BUFFER);
		}
		break;

	default:
		break;
	}

}



/*!
 * xvZ
 */
void rxSPH::calDensity(void)
{
	// MRK:calDensity
	RXREAL r0 = m_fInitDens;
	RXREAL h = m_fEffectiveRadius;

	for(uint i = 0; i < m_uNumParticles; ++i){
		Vec3 pos0;
		pos0[0] = m_hPos[4*i+0];
		pos0[1] = m_hPos[4*i+1];
		pos0[2] = m_hPos[4*i+2];

		m_hDens[i] = 0.0;

		// ߖTq
		for(vector<rxNeigh>::iterator itr = m_vNeighs[i].begin() ; itr != m_vNeighs[i].end(); ++itr){
			int j = itr->Idx;
			if(j < 0) continue;

			Vec3 pos1;
			pos1[0] = m_hPos[4*j+0];
			pos1[1] = m_hPos[4*j+1];
			pos1[2] = m_hPos[4*j+2];

			RXREAL r = sqrt(itr->Dist2);

			RXREAL q = h*h-r*r;

			m_hDens[i] += m_fMass*m_fWpoly6*q*q*q;
		}

		m_hPres[i] = m_fGasStiffness*(m_hDens[i]-r0);
	}

}


/*!
 * ͍CŠvZ
 */
void rxSPH::calForce(void)
{
	Vec3 rij, vji;
	RXREAL h = m_fEffectiveRadius;
	RXREAL r0 = m_fInitDens;
	

	// MRK:calForce
	for(uint i = 0; i < m_uNumParticles; ++i){
		m_hFrc[4*i+0] = 0.0;
		m_hFrc[4*i+1] = 0.0;
		m_hFrc[4*i+2] = 0.0;
		m_hFrc[4*i+3] = 0.0;

		m_hPres[i] = m_fGasStiffness*(m_hDens[i]-r0);
	}


	for(uint i = 0; i < m_uNumParticles; ++i){
		Vec3 pos0, vel0;
		pos0 = Vec3(m_hPos[4*i+0], m_hPos[4*i+1], m_hPos[4*i+2]);
		vel0 = Vec3(m_hVel[4*i+0], m_hVel[4*i+1], m_hVel[4*i+2]);

		RXREAL prsi = m_hPres[i]/(m_hDens[i]*m_hDens[i]);

		Vec3 pfrc(0.0), vfrc(0.0);
		for(vector<rxNeigh>::iterator itr = m_vNeighs[i].begin() ; itr != m_vNeighs[i].end(); ++itr){
			int j = itr->Idx;
			if(j < 0 || i == j) continue;

			Vec3 pos1, vel1;
			pos1 = Vec3(m_hPos[4*j+0], m_hPos[4*j+1], m_hPos[4*j+2]);
			vel1 = Vec3(m_hVel[4*j+0], m_hVel[4*j+1], m_hVel[4*j+2]);

			rij = pos0-pos1;
			vji = vel1-vel0;

			RXREAL r = norm(rij);//sqrt(itr->Dist2);
			RXREAL q = h-r;

			RXREAL prsj = m_hPres[j]/(m_hDens[j]*m_hDens[j]);

			Vec3 kernel;
			if(r <= 0.0){
				Vec3 rnd;
				rnd = RXFunc::Rand(Vec3(-0.0001), Vec3(0.0001));
				kernel = m_fGWspiky*rnd;
			}
			else{
				kernel = m_fGWspiky*q*q*rij/r;
			}

			// 
			pfrc += m_fMass*(prsi+prsj)*kernel;

			// S
			vfrc += m_fMass*(vji/m_hDens[j])*m_fLWvisc*q;
		}

		Vec3 force(0.0);
		force += -m_hDens[i]*pfrc;
		force += m_fViscosity*vfrc;

		// O͍
		if(m_fBuoyancy > 0){
			force += m_v3Gravity*m_fBuoyancy*(m_hDens[i]-r0);
		}
		else{
			force += m_v3Gravity*m_hDens[i];
		}

		m_hFrc[4*i+0] += force[0];
		m_hFrc[4*i+1] += force[1];
		m_hFrc[4*i+2] += force[2];
	}
}

/*!
 * @vZ
 */
void rxSPH::calNormal(void)
{
	// MRK:calNormal
	RXREAL h = m_fEffectiveRadius;

	for(uint i = 0; i < m_uNumParticles; ++i){
		Vec3 pos0;
		pos0[0] = m_hPos[4*i+0];
		pos0[1] = m_hPos[4*i+1];
		pos0[2] = m_hPos[4*i+2];

		Vec3 nrm(0.0);

		// ߖTq
		for(vector<rxNeigh>::iterator itr = m_vNeighs[i].begin() ; itr != m_vNeighs[i].end(); ++itr){
			int j = itr->Idx;
			if(j < 0 || i == j) continue;

			Vec3 pos1;
			pos1[0] = m_hPos[4*j+0];
			pos1[1] = m_hPos[4*j+1];
			pos1[2] = m_hPos[4*j+2];

			Vec3 rij = pos0-pos1;

			RXREAL r = sqrt(itr->Dist2);
			RXREAL q = h*h-r*r;

			nrm += (m_fMass/m_hDens[j])*m_fGWpoly6*q*q*rij;
		}

		//normalize(nrm);
		//nrm *= -1;

		m_hNrm[4*i+0] = nrm[0];
		m_hNrm[4*i+1] = nrm[1];
		m_hNrm[4*i+2] = nrm[2];
		m_hNrm[4*i+3] = 0.0;
	}

}


/*!
 * C/ƎOp`̌
 * @param[in] P0,P1 C/̒[_orC̓_
 * @param[in] V0,V1,V2 Op`̒_W
 * @param[out] I _W
 * @retval 1 _IŌ 
 * @retval 0 _Ȃ
 * @retval 2 Op`̕ʓ
 * @retval -1 Op`"degenerate"ł(ʐς0C܂C_ɂȂĂ)
 */
static int intersectSegmentTriangle(Vec3 P0, Vec3 P1,			// Segment
									Vec3 V0, Vec3 V1, Vec3 V2,	// Triangle
									Vec3 &I, Vec3 &n, float rp)			// Intersection point (return)
{
	// Op`̃GbWxNgƖ@
	Vec3 u = V1-V0;
	Vec3 v = V2-V0;
	n = Unit(cross(u, v));
	if(RXFunc::IsZeroVec(n)){
		return -1;	// Op`"degenerate"ł(ʐς0)
	}

	// 
	Vec3 dir = P1-P0;
	double a = dot(n, P0-V0);
	double b = dot(n, dir);
	if(fabs(b) < 1e-10){	// ƎOp`ʂs
		//if(fabs(a+b) < rp){
		//	I = P1+n*(rp-fabs(a+b));
		//	return 1;
		//}
		//else if(fabs(a) < rp){
		//	I = P1+n*(rp-fabs(a));
		//	return 1;
		//}
		//else{
			if(a == 0){
				return 2;	// ʏ
			}
			else{
				return 0;	// _Ȃ
			}
		//}
	}

	// _vZ

	// 2[_ꂼقȂʂɂ邩ǂ𔻒
	float r = -a/b;
	Vec3 offset = Vec3(0.0);
	float dn = 0;
	float sign_n = 1;
	//if(r < 0.0 || fabs(a) > fabs(b) || b > 0){
	//	return 0;
	//}
	if(a < 0){
		return 0;
	}

	if(r < 0.0){
		//if(fabs(a) < rp){
		//	r = 0.0;
		//	offset = -fabs(a)*n;

		//	dn = fabs(a);
		//}
		//else{
			return 0;
		//}
	}
	else{
		if(fabs(a) > fabs(b)){
			//if(fabs(a+b) < rp){
			//	r = 1.0;
			//	offset = -fabs(a+b)*n;

			//	dn = fabs(a+b);
			//}
			//else{
				return 0;
			//}
		}
		else{
			if(b > 0){
				return 0;
			}
		}
	}

	// ƕʂ̌_
	I = P0+r*dir;//+offset;

	// _Op`ɂ邩ǂ̔
	double uu, uv, vv, wu, wv, D;
	uu = dot(u, u);
	uv = dot(u, v);
	vv = dot(v, v);
	Vec3 w = I-V0;
	wu = dot(w, u);
	wv = dot(w, v);
	D = uv*uv-uu*vv;

	double s, t;
	s = (uv*wv-vv*wu)/D;
	if(s < 0.0 || s > 1.0){
		return 0;
	}
	
	t = (uv*wu-uu*wv)/D;
	if(t < 0.0 || (s+t) > 1.0){
		return 0;
	}

	//I -= offset;

	// a߂
	//Vec3 back = -dir*((rp-dn)/(dot(-dir, sign_n*n)));
	//I += back;

	return 1;
}

/*!
 * |SIuWFNgƂ̏Փ˔CՓˉ
 * @param[in] grid_hash Obh̃nbV
 * @param[in] pos0 OXebv̈ʒu
 * @param[inout] pos1 Vʒu
 * @param[inout] vel x
 * @param[in] dt ^CXebv
 * @return Փ˃IuWFNg̐
 */
int rxSPH::calCollisionPolygon(uint grid_hash, Vec3 &pos0, Vec3 &pos1, Vec3 &vel, RXREAL dt)
{
	int c = 0;

	uint start_index = m_hCellData.hPolyCellStart[grid_hash];
	if(start_index != 0xffffffff){	// ZłȂ̃`FbN
		uint end_index = m_hCellData.hPolyCellEnd[grid_hash];
		for(uint j = start_index; j < end_index; ++j){
			uint pidx = m_hCellData.hSortedPolyIdx[j];

			int vidx[3];
			vidx[0] = m_hTris[3*pidx+0];
			vidx[1] = m_hTris[3*pidx+1];
			vidx[2] = m_hTris[3*pidx+2];

			Vec3 vrts[3];
			vrts[0] = Vec3(m_hVrts[3*vidx[0]], m_hVrts[3*vidx[0]+1], m_hVrts[3*vidx[0]+2]);
			vrts[1] = Vec3(m_hVrts[3*vidx[1]], m_hVrts[3*vidx[1]+1], m_hVrts[3*vidx[1]+2]);
			vrts[2] = Vec3(m_hVrts[3*vidx[2]], m_hVrts[3*vidx[2]+1], m_hVrts[3*vidx[2]+2]);

			Vec3 cp, n;
			if(intersectSegmentTriangle(pos0, pos1, vrts[0], vrts[1], vrts[2], cp, n, m_fParticleRadius) == 1){
				double d = length(pos1-cp);
				n = Unit(n);

				RXREAL res = m_fRestitution;
				res = (res > 0) ? (res*fabs(d)/(dt*length(vel))) : 0.0f;
				Vec3 vr = -(1.0+res)*n*dot(n, vel);

				double l = norm(pos1-pos0);
				pos1 = cp+vr*(dt*d/l);
				vel += vr;

				c++;
			}
		}
	}

	return c;
}


/*!
 * ő̃IuWFNgƂ̏Փ˔CՓˉ
 * @param[in] pos0 OXebv̈ʒu
 * @param[inout] pos1 Vʒu
 * @param[inout] vel x
 * @param[in] dt ^CXebv
 * @return Փ˃IuWFNg̐
 */
int rxSPH::calCollisionSolid(Vec3 &pos0, Vec3 &pos1, Vec3 &vel, RXREAL dt)
{
	int c = 0;
	rxCollisionInfo coli;

	for(vector<rxSolid*>::iterator i = m_vSolids.begin(); i != m_vSolids.end(); ++i){
		if((*i)->GetDistanceR(pos1, m_fParticleRadius, coli)){
			RXREAL res = m_fDamping;
			res = (res > 0) ? (res*fabs(coli.Penetration())/(dt*norm(vel))) : 0.0f;
			vel -= (1+res)*dot(vel, coli.Normal())*coli.Normal();
			pos1 = coli.Contact();
		}
	}

	if(m_pBoundary->GetDistanceR(pos1, m_fParticleRadius, coli)){
		RXREAL res = m_fDamping;
		res = (res > 0) ? (res*fabs(coli.Penetration())/(dt*norm(vel))) : 0.0f;
		vel -= (1+res)*dot(vel, coli.Normal())*coli.Normal();
		pos1 = coli.Contact();
	}

	return c;
}

/*!
 * ʒuEx̍XV
 * @param[in] pos p[eBNʒu
 * @param[in] vel p[eBNx
 * @param[in] frc p[eBNɂ
 * @param[out] pos_new XVp[eBNʒu
 * @param[out] vel_new XVp[eBNx
 * @param[in] dt ^CXebv
 */
void rxSPH::integrate(const RXREAL *pos, const RXREAL *vel, const RXREAL *frc, 
					  RXREAL *pos_new, RXREAL *vel_new, RXREAL dt)
{
	for(uint i = 0; i < m_uNumParticles; ++i){
		Vec3 x, x_old, v, f, v_old;
		for(int k = 0; k < 3; ++k){
			x[k] = pos[DIM*i+k];
			v[k] = vel[DIM*i+k];
			f[k] = frc[DIM*i+k];
		}
		x_old = x;

		// Vxƈʒu
		v += dt*f/m_hDens[i];
		x += dt*v;

		// |SIuWFNgƂ̌
		if(m_iNumTris != 0){
			uint grid_hash0 = calGridHash(x_old);
			calCollisionPolygon(grid_hash0, x_old, x, v, dt);

			uint grid_hash1 = calGridHash(x);
			if(grid_hash1 != grid_hash0){
				calCollisionPolygon(grid_hash1, x_old, x, v, dt);
			}
		}

		// EƂ̏Փ˔
		calCollisionSolid(x_old, x, v, dt);

		// VxƈʒuōXV
		for(int k = 0; k < 3; ++k){
			pos_new[DIM*i+k] = x[k];
			vel_new[DIM*i+k] = v[k];
		}

	}
}



/*!
 * Z̗qߖTo
 * @param[in] pos TS
 * @param[in] prts p[eBNʒu
 * @param[in] gi,gj,gk ΏەZ
 * @param[out] neighs Tʊi[ߖTRei
 * @param[in] h La
 */
void rxSPH::getNeighborsInCell(Vec3 pos, RXREAL *prts, int gi, int gj, int gk, vector<rxNeigh> &neighs, RXREAL h)
{
	RXREAL h2 = h*h;

	uint grid_hash = calGridHash(gi, gj, gk);

	uint start_index = m_hCellData.hCellStart[grid_hash];
	if(start_index != 0xffffffff){	// ZłȂ̃`FbN
		uint end_index = m_hCellData.hCellEnd[grid_hash];
		for(uint j = start_index; j < end_index; ++j){
			uint idx = m_hCellData.hSortedIndex[j];

			Vec3 xij;
			xij[0] = pos[0]-prts[4*idx+0];
			xij[1] = pos[1]-prts[4*idx+1];
			xij[2] = pos[2]-prts[4*idx+2];

			rxNeigh neigh;
			neigh.Dist2 = norm2(xij);

			if(neigh.Dist2 <= h2){
				neigh.Idx = idx;
				//neigh.Dist = sqrt(neigh.Dist2);

				neighs.push_back(neigh);
			}
		}
	}
}

/*!
 * ߖTqT
 * @param[in] pos TS
 * @param[in] prts p[eBNʒu
 * @param[out] neighs Tʊi[ߖTRei
 * @param[in] h La
 */
void rxSPH::GetNearestNeighborsP(Vec3 pos, RXREAL *prts, vector<rxNeigh> &neighs, RXREAL h)
{
	if(h < 0.0) h = m_fEffectiveRadius;

	// ZCfbNX̎Zo
	int x = (pos[0]-m_v3EnvMin[0])/m_fCellWidth[0];
	int y = (pos[1]-m_v3EnvMin[1])/m_fCellWidth[1];
	int z = (pos[2]-m_v3EnvMin[2])/m_fCellWidth[2];

	int numArdGrid = (int)(h/m_fCellWidth[0])+1;
    for(int k = -numArdGrid; k <= numArdGrid; ++k){
        for(int j = -numArdGrid; j <= numArdGrid; ++j){
            for(int i = -numArdGrid; i <= numArdGrid; ++i){
				int i1 = x+i;
				int j1 = y+j;
				int k1 = z+k;
				if(i1 < 0 || i1 >= m_iGridSize[0] || j1 < 0 || j1 >= m_iGridSize[1] || k1 < 0 || k1 >= m_iGridSize[2]){
					continue;
				}

				getNeighborsInCell(pos, prts, i1, j1, k1, neighs, h);
			}
		}
	}
}

/*!
 * ߖTqT
 * @param[in] idx TSp[eBNCfbNX
 * @param[in] prts p[eBNʒu
 * @param[out] neighs Tʊi[ߖTRei
 * @param[in] h La
 */
void rxSPH::GetNearestNeighbors(int idx, RXREAL *prts, vector<rxNeigh> &neighs, RXREAL h)
{
	if(idx < 0 || idx >= (int)m_uNumParticles) return;

	Vec3 pos;
	pos[0] = prts[4*idx+0];
	pos[1] = prts[4*idx+1];
	pos[2] = prts[4*idx+2];

	GetNearestNeighborsP(pos, prts, neighs, h);
}

/*!
 * ߖTqT
 * @param[in] idx TSp[eBNCfbNX
 * @param[out] neighs Tʊi[ߖTRei
 * @param[in] h La
 */
void rxSPH::GetNearestNeighbors(int idx, vector<rxNeigh> &neighs, RXREAL h)
{
	GetNearestNeighbors(idx, m_hPos, neighs, h);
}


/*!
 * ߖTqT()
 * @param[in] pos TS
 * @param[in] prts p[eBNʒu
 * @param[out] neighs Tʊi[ߖTRei
 * @param[in] h La
 */
void rxSPH::GetNearestNeighborsDirectP(Vec3 pos0, RXREAL *prts, vector<rxNeigh> &neighs, RXREAL h)
{
	if(h < 0.0) h = m_fEffectiveRadius;
	RXREAL h2 = h*h;

	for(int i = 0; i < (int)m_uNumParticles; i++){
		Vec3 pos1;
		pos1[0] = prts[4*i+0];
		pos1[1] = prts[4*i+1];
		pos1[2] = prts[4*i+2];

		rxNeigh neigh;
		neigh.Dist2 = (RXREAL)norm2(pos0-pos1);

		if(neigh.Dist2 <= h2 && neigh.Dist2 > RX_FEQ_EPS){
			neigh.Idx = i;
			neighs.push_back(neigh);
		}
	}
}

/*!
 * ߖTqT()
 * @param[in] idx TSp[eBNCfbNX
 * @param[in] prts p[eBNʒu
 * @param[out] neighs Tʊi[ߖTRei
 * @param[in] h La
 */
void rxSPH::GetNearestNeighborsDirect(int idx, RXREAL *prts, vector<rxNeigh> &neighs, RXREAL h)
{
	if(idx < 0 || idx >= (int)m_uNumParticles) return;

	Vec3 pos;
	pos[0] = prts[4*idx+0];
	pos[1] = prts[4*idx+1];
	pos[2] = prts[4*idx+2];

	GetNearestNeighborsDirectP(pos, prts, neighs, h);
}

/*!
 * ߖTqT()
 * @param[in] idx TSp[eBNCfbNX
 * @param[out] neighs Tʊi[ߖTRei
 * @param[in] h La
 */
void rxSPH::GetNearestNeighborsDirect(int idx, vector<rxNeigh> &neighs, RXREAL h)
{
	GetNearestNeighborsDirect(idx, m_hPos, neighs, h);
}

/*!
 * ׂẴp[eBN̋ߖTT
 * @param[in] prts p[eBNʒu
 * @param[in] h La
 */
void rxSPH::searchNeighbors(RXREAL *prts, RXREAL h)
{
	// MRK:searchNeighbors
	assert(m_bInitialized);

	if(h < 0.0) h = m_fEffectiveRadius;

	for(int i = 0; i < (int)m_uNumParticles; i++){
		m_vNeighs[i].clear();
		//GetNearestNeighborsDirect(i, prts, m_vNeighs[i], h);
		GetNearestNeighbors(i, prts, m_vNeighs[i], h);
	}
}

/*!
 * ObhnbVľvZ
 * @param[in] x,y,z Obhʒu
 * @return ObhnbVl
 */
uint rxSPH::calGridHash(int x, int y, int z)
{
	x = (x < 0 ? 0 : (x >= m_iGridSize[0] ? m_iGridSize[0]-1 : x));
	y = (y < 0 ? 0 : (y >= m_iGridSize[1] ? m_iGridSize[1]-1 : y));
	z = (z < 0 ? 0 : (z >= m_iGridSize[2] ? m_iGridSize[2]-1 : z));
	return z*m_iGridSize[1]*m_iGridSize[0]+y*m_iGridSize[0]+x;
}
/*!
 * ObhnbVľvZ
 * @param[in] pos p[eBNW
 * @return ObhnbVl
 */
uint rxSPH::calGridHash(Vec3 pos)
{
	pos -= m_v3EnvMin;

	// ZCfbNX̎Zo
	int x = pos[0]/m_fCellWidth[0];
	int y = pos[1]/m_fCellWidth[1];
	int z = pos[2]/m_fCellWidth[2];
	return calGridHash(x, y, z);
}

/*!
 * p[eBN𕪊ZɊi[
 *  - p[eBN̑ObhnbVvZĊi[
 * @param[in] pidx p[eBNCfbNX
 * @param[in] p Sp[eBN̍Wi[z
 */
void rxSPH::setObjectToCell(int pidx, RXREAL *p)
{
	Vec3 pos;
	pos[0] = p[4*pidx+0];
	pos[1] = p[4*pidx+1];
	pos[2] = p[4*pidx+2];

	// nbVlvZ
	uint hash = calGridHash(pos);

	m_hCellData.hSortedIndex[pidx] = pidx;
	m_hCellData.hGridParticleHash[pidx] = hash;
}

/*!
 * Sp[eBN𕪊ZɊi[
 *  - ep[eBN̑ObhnbVvZĊi[
 * @param[in] p Sp[eBN̍Wi[z
 */
void rxSPH::setObjectToCell(RXREAL *p)
{
	int mem_size1 = m_uNumParticles*sizeof(uint);
	int mem_size2 = m_hCellData.uNumCells*sizeof(uint);
	memset(m_hCellData.hSortedIndex, 0, mem_size1);
	memset(m_hCellData.hGridParticleHash, 0, mem_size1);
	memset(m_hCellData.hCellStart, 0xffffffff, mem_size2);
	memset(m_hCellData.hCellEnd, 0xffffffff, mem_size2);

	if(m_uNumParticles == 0) return;

	// ep[eBÑObhnbV̌vZ
	for(int i = 0; i < (int)m_uNumParticles; ++i){
		setObjectToCell(i, p);
	}

	// ObhnbVŃ\[g
	vector<rxHashSort> hash_and_value;
	hash_and_value.resize(m_uNumParticles);
	for(int i = 0; i < (int)m_uNumParticles; ++i){
		hash_and_value[i].hash = m_hCellData.hGridParticleHash[i];
		hash_and_value[i].value  = m_hCellData.hSortedIndex[i];
	}
	std::sort(hash_and_value.begin(), hash_and_value.end(), LessHash);
	for(int i = 0; i < (int)m_uNumParticles; ++i){
		m_hCellData.hSortedIndex[i] = hash_and_value[i].value;
		m_hCellData.hGridParticleHash[i] = hash_and_value[i].hash;
	}

	// p[eBNz\[gꂽԂɕёւC
	// eZ̎n܂ƏĨCfbNX
	for(int i = 0; i < (int)m_uNumParticles; i++){
		int hash = m_hCellData.hGridParticleHash[i];

		if(i == 0){
			m_hCellData.hCellStart[hash] = i;
			m_hCellData.hCellEnd[hash] = i;
		}
		else{
			int prev_hash = m_hCellData.hGridParticleHash[i-1];

			if(i == 0 || hash != prev_hash){
				m_hCellData.hCellStart[hash] = i;
				if(i > 0){
					m_hCellData.hCellEnd[prev_hash] = i;
				}
			}

			if(i == m_uNumParticles-1){
				m_hCellData.hCellEnd[hash] = i+1;
			}
		}
	}
}

/*!
 * ZɊi[ꂽ|S擾
 * @param[in] gi,gj,gk ΏەZ
 * @param[out] polys |S
 * @return i[|S
 */
int rxSPH::GetPolygonsInCell(int gi, int gj, int gk, vector<int> &polys)
{
	uint grid_hash = calGridHash(gi, gj, gk);

	uint start_index = m_hCellData.hPolyCellStart[grid_hash];
	if(start_index != 0xffffffff){	// ZłȂ̃`FbN

		int cnt = 0;
		uint end_index = m_hCellData.hPolyCellEnd[grid_hash];
		for(uint j = start_index; j < end_index; ++j){
			uint idx = m_hCellData.hSortedPolyIdx[j];

			polys.push_back(idx);

			cnt++;
		}

		return cnt;
	}

	return 0;
}
/*!
 * Z̃|S̗L𒲂ׂ
 * @param[in] gi,gj,gk ΏەZ
 * @return |Si[Ătrue
 */
bool rxSPH::IsPolygonsInCell(int gi, int gj, int gk)
{
	uint grid_hash = calGridHash(gi, gj, gk);

	uint start_index = m_hCellData.hPolyCellStart[grid_hash];
	if(start_index != 0xffffffff){	// ZłȂ̃`FbN

		int cnt = 0;
		uint end_index = m_hCellData.hPolyCellEnd[grid_hash];
		for(uint j = start_index; j < end_index; ++j){
			uint idx = m_hCellData.hSortedPolyIdx[j];
			cnt++;
			break;
		}

		return (cnt > 0);
	}

	return false;
}


/*!
 * Sp[eBN𕪊ZɊi[
 */
void rxSPH::SetParticlesToCell(void)
{
	// Zɗqo^
	setObjectToCell(m_hPos);

	// ߖTqT
	searchNeighbors(m_hPos, m_fEffectiveRadius);
}

/*!
 * ԕ@̏
 * @param[in] vMin ̍ŏW
 * @param[in] vMax ̍őW
 * @param[in] h ea
 */
void rxSPH::setupCells(const Vec3 &vMin, const Vec3 &vMax, const double &h)
{
	if(h < RX_EPS) return;

	Vec3 world_size = vMax-vMin;
	Vec3 world_origin = vMin;

	double max_axis = RXFunc::Max3(world_size);

	int d = (int)(log(max_axis/h)/log(2.0)+0.5);
	int n = (int)(pow(2.0, (double)d)+0.5);
	double cell_width = max_axis/n;

	d = (int)(log(world_size[0]/cell_width)/log(2.0)+0.5);
	m_iGridSize[0] = (int)(pow(2.0, (double)d)+0.5);
	d = (int)(log(world_size[1]/cell_width)/log(2.0)+0.5);
	m_iGridSize[1] = (int)(pow(2.0, (double)d)+0.5);;
	d = (int)(log(world_size[2]/cell_width)/log(2.0)+0.5);
	m_iGridSize[2] = (int)(pow(2.0, (double)d)+0.5);;

	m_fCellWidth[0] = cell_width;
	m_fCellWidth[1] = cell_width;
	m_fCellWidth[2] = cell_width;

	m_hCellData.uNumCells = m_iGridSize[0]*m_iGridSize[1]*m_iGridSize[2];
}

/*!
 * |S𕪊ZɊi[
 * @param[in] vrts |S_
 * @param[in] nv _
 * @param[in] tris bV
 * @param[in] nt bV
 */
void rxSPH::setPolysToCell(RXREAL *vrts, int nv, int* tris, int nt)
{
	// MRK:setPolysToCell
	int mem_size2 = m_hCellData.uNumCells*sizeof(uint);
	memset(m_hCellData.hPolyCellStart, 0xffffffff, mem_size2);
	memset(m_hCellData.hPolyCellEnd, 0, mem_size2);

	int num_hash = 0;

	// ep[eBÑObhnbV̌vZ
	vector<uint> tri_hash, tri_idx;
	vector<Vec3> tri_vrts, tri_vrts_c;
	tri_vrts.resize(3);
	tri_vrts_c.resize(3);
	for(int i = 0; i < nt; i++){
		for(int j = 0; j < 3; ++j){
			Vec3 pos;
			pos[0] = vrts[3*tris[3*i+j]+0];
			pos[1] = vrts[3*tris[3*i+j]+1];
			pos[2] = vrts[3*tris[3*i+j]+2];
			tri_vrts[j] = pos;
		}

		Vec3 nrm = Unit(cross(tri_vrts[1]-tri_vrts[0], tri_vrts[2]-tri_vrts[0]));

		// |SBBox
		Vec3 bmin, bmax;
		bmin = tri_vrts[0];
		bmax = tri_vrts[0];
		for(int j = 1; j < 3; ++j){
			for(int k = 0; k < 3; ++k){
				if(tri_vrts[j][k] < bmin[k]) bmin[k] = tri_vrts[j][k];
				if(tri_vrts[j][k] > bmax[k]) bmax[k] = tri_vrts[j][k];
			}
		}

		// BBoxƏdȂZ
		bmin -= m_v3EnvMin;
		bmax -= m_v3EnvMin;

		// ZCfbNX̎Zo
		int bmin_gidx[3], bmax_gidx[3];
		for(int k = 0; k < 3; ++k){
			bmin_gidx[k] = bmin[k]/m_fCellWidth[k];
			bmax_gidx[k] = bmax[k]/m_fCellWidth[k];

			bmin_gidx[k] = RX_CLAMP(bmin_gidx[k], 0, m_iGridSize[k]-1);
			bmax_gidx[k] = RX_CLAMP(bmax_gidx[k], 0, m_iGridSize[k]-1);
		}

		// eZɃ|S܂܂邩`FbN
		Vec3 len = Vec3(m_fCellWidth[0], m_fCellWidth[1], m_fCellWidth[2]);
		Vec3 cen(0.0);
		for(int x = bmin_gidx[0]; x <= bmax_gidx[0]; ++x){
			for(int y = bmin_gidx[1]; y <= bmax_gidx[1]; ++y){
				for(int z = bmin_gidx[2]; z <= bmax_gidx[2]; ++z){
					cen = m_v3EnvMin+Vec3(x+0.5, y+0.5, z+0.5)*len;

					for(int j = 0; j < 3; ++j){
						tri_vrts_c[j] = (tri_vrts[j]-cen)/len;
					}

					if(RXFunc::polygon_intersects_cube(tri_vrts_c, nrm, 0)){
						// nbVlvZ
						uint hash = calGridHash(x, y, z);

						tri_idx.push_back((uint)i);
						tri_hash.push_back(hash);

						//m_hCellData.hPolyCellStart[hash] = 0;

						num_hash++;
					}
				}
			}
		}
	}

	RXCOUT << "polygon hash : " << num_hash << endl;

	m_hCellData.uNumPolyHash = (uint)num_hash;

	int mem_size1 = m_hCellData.uNumPolyHash*sizeof(uint);
	m_hCellData.hSortedPolyIdx = new uint[m_hCellData.uNumPolyHash];
	m_hCellData.hGridPolyHash = new uint[m_hCellData.uNumPolyHash];
	memcpy(m_hCellData.hSortedPolyIdx, &tri_idx[0], mem_size1);
	memcpy(m_hCellData.hGridPolyHash, &tri_hash[0], mem_size1);

	// ObhnbVŃ\[g
	vector<rxHashSort> hash_and_value;
	hash_and_value.resize(m_hCellData.uNumPolyHash);
	for(int i = 0; i < (int)m_hCellData.uNumPolyHash; ++i){
		hash_and_value[i].hash  = m_hCellData.hGridPolyHash[i];
		hash_and_value[i].value = m_hCellData.hSortedPolyIdx[i];
	}
	std::sort(hash_and_value.begin(), hash_and_value.end(), LessHash);
	for(int i = 0; i < (int)m_hCellData.uNumPolyHash; ++i){
		m_hCellData.hSortedPolyIdx[i] = hash_and_value[i].value;
		m_hCellData.hGridPolyHash[i]  = hash_and_value[i].hash;
	}

	// p[eBNz\[gꂽԂɕёւC
	// eZ̎n܂ƏĨCfbNX
	for(uint i = 0; i < m_hCellData.uNumPolyHash; ++i){
		uint hash = m_hCellData.hGridPolyHash[i];

		if(i == 0){
			m_hCellData.hPolyCellStart[hash] = i;
		}
		else{
			uint prev_hash = m_hCellData.hGridPolyHash[i-1];

			if(i == 0 || hash != prev_hash){
				m_hCellData.hPolyCellStart[hash] = i;
				if(i > 0){
					m_hCellData.hPolyCellEnd[prev_hash] = i;
				}
			}

			if(i == m_uNumParticles-1){
				m_hCellData.hPolyCellEnd[hash] = i+1;
			}
		}
	}
}

/*!
 * |S𕪊ZɊi[
 */
void rxSPH::SetPolygonsToCell(void)
{
	setPolysToCell(m_hVrts, m_iNumVrts, m_hTris, m_iNumTris);
}


/*!
 * TpZ̕`
 * @param[in] i,j,k Obh̃CfbNX
 */
void rxSPH::DrawCell(int i, int j, int k)
{
	glPushMatrix();
	glTranslated(m_v3EnvMin[0], m_v3EnvMin[1], m_v3EnvMin[2]);
	glTranslatef((i+0.5)*m_fCellWidth[0], (j+0.5)*m_fCellWidth[1], (k+0.5)*m_fCellWidth[2]);
	glutWireCube(m_fCellWidth[0]);
	glPopMatrix();
}

/*!
 * TpObh̕`
 * @param[in] col p[eBN܂܂Z̐F
 * @param[in] col2 |S܂܂Z̐F
 * @param[in] sel _ɑIꂽẐݕ`(1ŐVZIC2łłɑIĂZ`C0łׂẴZ`)
 */
void rxSPH::DrawCells(Vec3 col, Vec3 col2, int sel)
{
	glPushMatrix();

	if(sel){
		// _ɑI񂾃ZƂ̒̃p[eBN̂ݕ`
		static int grid_hash = 0;
		static uint start_index = 0xffffffff;
		if(sel == 1){
			do{
				grid_hash = RXFunc::Nrand(m_hCellData.uNumCells-1);
				start_index = m_hCellData.hCellStart[grid_hash];
			}while(start_index == 0xffffffff);
		}

		uint w = grid_hash%(m_iGridSize[0]*m_iGridSize[1]);
		DrawCell(w%m_iGridSize[0], w/m_iGridSize[0], grid_hash/(m_iGridSize[0]*m_iGridSize[1]));

		glColor3d(1.0, 0.0, 0.0);
		glPointSize(10.0);
		glBegin(GL_POINTS);

		int c = 0;
		uint end_index = m_hCellData.hCellEnd[grid_hash];
		for(uint j = start_index; j < end_index; ++j){
			uint idx = m_hCellData.hSortedIndex[j];
			Vec3 pos;
			pos[0] = m_hPos[4*idx+0];
			pos[1] = m_hPos[4*idx+1];
			pos[2] = m_hPos[4*idx+2];
			
			glVertex3dv(pos);

			c++;
		}
		glEnd();

		cout << "cell(" << grid_hash << ") : " << c << endl;
	}
	else{
		// p[eBN or |S܂ޑSZ̕`
		RXFOR3(0, m_iGridSize[0], 0, m_iGridSize[1], 0, m_iGridSize[2]){
			bool disp = false;
			uint grid_hash = calGridHash(i, j, k);
			uint start_index = m_hCellData.hCellStart[grid_hash];
			uint start_index_poly = 0xffffffff;
		
			if(m_hCellData.uNumPolyHash) start_index_poly = m_hCellData.hPolyCellStart[grid_hash];

			if(start_index != 0xffffffff){
				glColor3dv(col2.data);
				disp = true;
			}
			if(start_index_poly != 0xffffffff){
				glColor3dv(col.data);
				disp = true;
			}

			if(disp){
				DrawCell(i, j, k);
			}
		}
	}

	glPopMatrix();
}

/*!
 * ő̏Q̕`
 */
void rxSPH::DrawObstacles(void)
{
	for(vector<rxSolid*>::iterator i = m_vSolids.begin(); i != m_vSolids.end(); ++i){
		(*i)->Draw();
	}
}




/*!
 * Op`|SɂQ
 * @param[in] vrts _
 * @param[in] tris bV
 */
void rxSPH::SetPolygonObstacle(const vector<Vec3> &vrts, const vector<Vec3> &nrms, const vector< vector<int> > &tris)
{
	int vn = (int)vrts.size();
	int n = (int)tris.size();

	if(m_hVrts) delete [] m_hVrts;
	if(m_hTris) delete [] m_hTris;
	m_hVrts = new RXREAL[vn*3];
	m_hTris = new int[n*3];

	for(int i = 0; i < vn; ++i){
		for(int j = 0; j < 3; ++j){
			m_hVrts[3*i+j] = vrts[i][j];
		}
	}

	for(int i = 0; i < n; ++i){
		for(int j = 0; j < 3; ++j){
			m_hTris[3*i+j] = tris[i][j];
		}
	}

	m_iNumVrts = vn;
	m_iNumTris = n;
	RXCOUT << "the number of triangles : " << m_iNumTris << endl;

	//SetParticlesToCell();
	setPolysToCell(m_hVrts, m_iNumVrts, m_hTris, m_iNumTris);
}

/*!
 * {bNX^Q
 * @param[in] cen {bNXSW
 * @param[in] ext {bNX̑傫(ӂ̒1/2)
 * @param[in] ang {bNX̊px(IC[p)
 * @param[in] flg L/tO
 */
void rxSPH::SetBoxObstacle(Vec3 cen, Vec3 ext, Vec3 ang, int flg)
{
	rxSolidBox *box = new rxSolidBox(cen-ext, cen+ext, 1);
	double m[16];
	EulerToMatrix(m, ang[0], ang[1], ang[2]);
	box->SetMatrix(m);

	m_vSolids.push_back(box);

}

/*!
 * ^Q
 * @param[in] cen ̒SW
 * @param[in] rad ̂̔a
 * @param[in] flg L/tO
 */
void rxSPH::SetSphereObstacle(Vec3 cen, double rad, int flg)
{
	rxSolidSphere *sphere = new rxSolidSphere(cen, rad, 1);
	m_vSolids.push_back(sphere);
}

/*!
 * ^Q𓮂
 * @param[in] b ̔ԍ
 * @param[in] disp ړ
 */
void rxSPH::MoveSphereObstacle(int b, Vec3 disp)
{
	if(m_vSolids.empty()) return;
	m_vSolids[b]->SetPosition(m_vSolids[b]->GetPosition()+disp);
}

/*!
 * ^Q̈ʒu擾
 * @param[in] b ̔ԍ
 */
Vec3 rxSPH::GetSphereObstaclePos(int b)
{
	if(m_vSolids.empty()) return Vec3(0.0);
	return m_vSolids[b]->GetPosition();
}


/*!
 * VBOzXgփf[^]C擾
 * @param[in] type f[^̎
 * @return zXg̃f[^
 */
RXREAL* rxSPH::GetArrayVBO(rxParticleArray type, bool d2h)
{
	assert(m_bInitialized);
 
	RXREAL* hdata = 0;

	unsigned int vbo = 0;

	switch(type){
	default:
	case RX_POSITION:
		hdata = m_hPos;
		vbo = m_posVBO;
		break;

	case RX_VELOCITY:
		hdata = m_hVel;
		break;

	case RX_DENSITY:
		hdata = m_hDens;
		break;

	case RX_PRESSURE:
		hdata = m_hPres;
		break;

	case RX_NORMAL:
		hdata = m_hNrm;
		break;

	case RX_FORCE:
		hdata = m_hFrc;
		break;

	case RX_UPDATED_POSITION:
		hdata = m_hUpPos;
		break;

	case RX_EIGEN_VALUE:
		hdata = m_hEigen;
		break;

	case RX_ROTATION_MATRIX:
		hdata = m_hRMatrix;
		break;

	case RX_TRANSFORMATION:
		hdata = m_hG;
		break;
	}

	return hdata;
}

/*!
 * zXgVBOփf[^]
 * @param[in] type f[^̎
 * @param[in] data zXg̃f[^
 * @param[in] start f[^̊JnCfbNX
 * @param[in] count ǉ
 */
void rxSPH::SetArrayVBO(rxParticleArray type, const RXREAL* data, int start, int count)
{
	assert(m_bInitialized);
 
	switch(type){
	default:
	case RX_POSITION:
		{
			if(m_bUseOpenGL){
				glBindBuffer(GL_ARRAY_BUFFER, m_posVBO);
				glBufferSubData(GL_ARRAY_BUFFER, start*4*sizeof(RXREAL), count*4*sizeof(RXREAL), data);
				glBindBuffer(GL_ARRAY_BUFFER, 0);
			}
		}
		break;

	case RX_VELOCITY:
		//CuCopyArrayToDevice(m_dVel, data, start*DIM*sizeof(RXREAL), count*DIM*sizeof(RXREAL));
		break;

	case RX_NORMAL:
		//CuCopyArrayToDevice(m_dNrm, data, start*DIM*sizeof(RXREAL), count*DIM*sizeof(RXREAL));
		break;
	}	   
}


/*!
 * J[tB[hlvZ
 * @param[in] pos vZʒu
 * @return J[tB[hl
 */
double rxSPH::CalColorField(double x, double y, double z)
{
	// MRK:CalColorField
	RXREAL c = 0.0;
	Vec3 pos(x, y, z);

	if(pos[0] < m_v3EnvMin[0]) return c;
	if(pos[0] > m_v3EnvMax[0]) return c;
	if(pos[1] < m_v3EnvMin[1]) return c;
	if(pos[1] > m_v3EnvMax[1]) return c;
	if(pos[2] < m_v3EnvMin[2]) return c;
	if(pos[2] > m_v3EnvMax[2]) return c;

	RXREAL h = m_fEffectiveRadius;

	vector<rxNeigh> ne;
	GetNearestNeighborsP(pos, m_hPos, ne, h);

	// ߖTq
	for(vector<rxNeigh>::iterator itr = ne.begin(); itr != ne.end(); ++itr){
		int j = itr->Idx;
		if(j < 0) continue;

		Vec3 pos1;
		pos1[0] = m_hPos[DIM*j+0];
		pos1[1] = m_hPos[DIM*j+1];
		pos1[2] = m_hPos[DIM*j+2];

		RXREAL r = sqrt(itr->Dist2);

		RXREAL q = h*h-r*r;
		c += m_fMass*m_fWpoly6*q*q*q;

		//RXREAL q = r/h;
		//if(q <= 1.0 && q >= 0.0){
		//	c += m_fMass*(1.0-q)*(1.0-q);
		//}
	}

	return c;
}



//-----------------------------------------------------------------------------
// MARK:A֐l
//-----------------------------------------------------------------------------
double rxSPH::GetImplicit(double x, double y, double z)
{
	return CalColorField(x, y, z);
}

/*!
 * p[eBNObh̉A֐lvZ
 * @param[in] pnx,pny,pnz Obh̎w nx=2^pnx
 * @param[in] minp Obh̍ŏW
 * @param[in] d Obh
 * @param[out] hF A֐l(nx~ny~nz̔z)
 */
void rxSPH::CalImplicitField(int n[3], Vec3 minp, Vec3 d, RXREAL *hF)
{
	int slice0 = n[0];
	int slice1 = n[0]*n[1];

	for(int k = 0; k < n[2]; ++k){
		for(int j = 0; j < n[1]; ++j){
			for(int i = 0; i < n[0]; ++i){
				int idx = k*slice1+j*slice0+i;
				Vec3 pos = minp+Vec3(i, j, k)*d;
				hF[idx] = GetImplicit(pos[0], pos[1], pos[2]);
			}
		}
	}
}

/*!
 * p[eBNObh̉A֐lvZ
 * @param[in] pnx,pny,pnz Obh̎w nx=2^pnx
 * @param[in] minp Obh̍ŏW
 * @param[in] d Obh
 * @param[out] hF A֐l(nx~ny~nz̔z)
 */
void rxSPH::CalImplicitFieldDevice(int n[3], Vec3 minp, Vec3 d, RXREAL *dF)
{
	RXREAL *hF = new RXREAL[n[0]*n[1]*n[2]];

	CalImplicitField(n, minp, d, hF);
	CuCopyArrayToDevice(dF, hF, 0, n[0]*n[1]*n[2]*sizeof(RXREAL));

	delete [] hF;
}





//-----------------------------------------------------------------------------
// MARK:V~f[^̏o
//-----------------------------------------------------------------------------
/*!
 * V~[Vݒ(p[eBNĆ͈CxCʂȂ)
 * @param[in] fn o̓t@C
 */
void rxSPH::OutputSetting(string fn)
{
	ofstream fout;
	fout.open(fn.c_str());
	if(!fout){
		RXCOUT << fn << " couldn't open." << endl;
		return;
	}

	fout << m_uNumParticles << endl;
	fout << m_v3EnvMin[0] << " " << m_v3EnvMin[1] << " " << m_v3EnvMin[2] << endl;
	fout << m_v3EnvMax[0] << " " << m_v3EnvMax[1] << " " << m_v3EnvMax[2] << endl;
	fout << m_fInitDens << endl;
	fout << m_fMass << endl;
	fout << m_iKernelParticles << endl;

	fout.close();
}

